/* ops.c - op tables
 * Rob Earhart
 * $Id$
 */
/***********************************************************
        Copyright 1997 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include <stdio.h>
#include "filter.h"
#include "y.tab.h"
#include "msg.h"

static const char rcsid[] = "$Id$";

static void Strings_Print(Base *base, FILE *stream) {
  Strings *it = (Strings *)base;
  fprintf(stream, "\"%s\"", it->str);
  if (it->list.next) {
    putc(' ', stream);
    base_Print(it->list.next, stream);
  }
}

static void Tests_Print(Base *base, FILE *stream) {
  Tests *it = (Tests *)base;
  base_Print(it->test, stream);
  if (it->list.next) {
    fputs(", ", stream);
    base_Print(it->list.next, stream);
  }
}

static int Tests_Eval(Test *test) {
  Tests *it = (Tests *)test;
  fputs("Tests_Eval called: this should never happen :-)\n", stderr);
  exit(1);
  return 0;			/* stupid compilers :-) */
}

static void All_Of_Print(Base *base, FILE *stream) {
  All_Of *it = (All_Of *)base;
  fputs("all-of (", stream);
  base_Print(it->tests, stream);
  putc(')', stream);
}

static int All_Of_Eval(Test *test) {
  All_Of *it = (All_Of *)test;
  Tests *tests;
  for (tests = it->tests; tests; tests = (Tests *)tests->list.next)
    if (! test_Eval(tests->test))
      return 0;
  return 1;
}

static void Any_Of_Print(Base *base, FILE *stream) {
  Any_Of *it = (Any_Of *)base;
  fputs("any-of (", stream);
  base_Print(it->tests, stream);
  putc(')', stream);
}

static int Any_Of_Eval(Test *test) {
  Any_Of *it = (Any_Of *)test;
  Tests *tests;
  for (tests = it->tests; tests; tests = (Tests *)tests->list.next)
    if (test_Eval(tests->test))
      return 1;
  return 0;
}

static void Exists_Print(Base *base, FILE *stream) {
  Exists *it = (Exists *)base;
  fputs("exists (", stream);
  base_Print(it->headers, stream);
  putc(')', stream);
}

static int Exists_Eval(Test *test) {
  Exists *it = (Exists *)test;
  Strings *header;
  for (header = it->headers; header; header=(Strings *)header->list.next)
    if (! (msg_Header(header->str)->name)) return 0;
  return 1;
}

static void False_Print(Base *base, FILE *stream) {
  fputs("false", stream);
}

static int False_Eval(Test *test) {
  return 0;
}

static void Header_Print(Base *base, FILE *stream) {
  Header *it = (Header *)base;
  fputs("header (", stream);
  base_Print(it->headers, stream);
  fputs(") ", stream);
  switch(it->matcher) {
  case CONTAINS:
    fputs("contains", stream);
    break;
  case CONTAINS_NOCASE:
    fputs("contains-nocase", stream);
    break;
  case IS:
    fputs("is", stream);
    break;
  case IS_NOCASE:
    fputs("is-nocase", stream);
    break;
  case MATCHES:
    fputs("matches", stream);
    break;
  case MATCHES_NOCASE:
    fputs("matches-nocase", stream);
    break;
  default:
    fputs("(unknown)", stream);
  }
  fputs(" (", stream);
  base_Print(it->keys, stream);
  putc(')', stream);
}

static int contains(const char *text,
		    const char *what) {
  /* Cheap.  Might be efficient, tho. */
  return strstr(text, what);
}

static int contains_nocase(const char *text,
			   const char *what) {
  /* There's an efficient algorithm for this,
   * but this isn't it.  (Ideally, all these functions should
   * be replaced with parse-time translations into regexps... XXX */
  /* It's probably not worth duplicating, recasing, and
   * strstr'ing, tho... */
  unsigned len = strlen(what);
  if (! what[0]) return 1;
  for (; *text; text++)
    if ((isupper(*text) ? tolower(*text) : *text) ==
	(isupper(*what) ? tolower(*what) : *what)
	&& ! strncasecmp(text, what, len))
      return 1;
  return 0;
}

static int is(const char *text,
	      const char *what) {
  return ! strcmp(text, what);
}

static int is_nocase(const char *text,
		     const char *what) {
  return ! strcasecmp(text, what);
}

/* Again, there's an efficient algorithm for this, but this isn't it,
 * but this works.  Kinda.  (It's slow) (The real problem is that you
 * pay a hefty price on *s, which use backtracking to do their dirty
 * work).  And because it's a bad algorithm, I haven't taken the time
 * to optimize the code - I think this is pretty clear, and I'd rather
 * not take the time and effort to generate something hard to read
 * when the algorithm sucks in the first place.  */

static int match(const char *text,
		 const char *what,
		 int nocase) {
  switch (*what) {
  case '\0':
    return (*text == '\0' ? 1 : 0);
  case '?':
    if (! *text) return 0;
    return match(text+1, what+1, nocase);
  case '*':
    for(; *text; text++)
      if (match(text, what+1, nocase))
	return 1;
    return 0;
  case '\\':
    what++;
    /* fall-through */
  default:
    if ((nocase && isupper(*text) ? tolower(*text) : *text) ==
	(nocase && isupper(*what) ? tolower(*what) : *what))
      return match(text+1, what+1, nocase);
    else
      return 0;
  }
}

static int matches(const char *text,
		   const char *what) {
  return match(text, what, 0);
}

static int matches_nocase(const char *text,
			  const char *what) {
  return match(text, what, 1);
}

static int Header_Eval(Test *test) {
  Header *it = (Header *)test;
  Strings *header;
  Strings *key;
  MsgHeader *mh;

  for (header = it->headers; header; header = (Strings *)header->list.next)
    for (mh = msg_Header(header->str); mh->name; mh++)
      for (key = it->keys; key; key = (Strings *)key->list.next)
	switch (it->matcher) {
	case CONTAINS:
	  if (contains(mh->body, key->str))
	    return 1;
	  break;
	case CONTAINS_NOCASE:
	  if (contains_nocase(mh->body, key->str))
	    return 1;
	  break;
	case IS:
	  if (is(mh->body, key->str))
	    return 1;
	  break;
	case IS_NOCASE:
	  if (is_nocase(mh->body, key->str))
	    return 1;
	  break;
	case MATCHES:
	  if (matches(mh->body, key->str))
	    return 1;
	  break;
	case MATCHES_NOCASE:
	  if (matches_nocase(mh->body, key->str))
	    return 1;
	  break;
	default:
	  return 0;
	}

  return 0;
}

static void Not_Print(Base *base, FILE *stream) {
  Not *it = (Not *)base;
  fputs("not ", stream);
  base_Print(it->it, stream);
}

static int Not_Eval(Test *test) {
  Not *it = (Not *)test;
  return ! test_Eval(it->it);
}

static void Size_Print(Base *base, FILE *stream) {
  Size *it = (Size *)base;
  fprintf(stream, "size %s %u",
	  (it->over ? "over" : "under"),
	  it->value);
}

static int Size_Eval(Test *test) {
  Size *it = (Size *)test;
  return (it->over ? msg_Size() > it->value : msg_Size() < it->value);
}

static void Support_Print(Base *base, FILE *stream) {
  Support *it = (Support *)base;
  fputs("support ", stream);
  putc('"', stream);
  fputs(it->extension, stream);
  putc('"', stream);
}

static int Support_Eval(Test *test) {
  return 0;			/* add extension mechanism... XXX */
}

static void True_Print(Base *base, FILE *stream) {
  fputs("true", stream);
}

static int True_Eval(Test *test) {
  return 1;
}

static void Clause_Print(Base *base, FILE *stream) {
  Clause *it = (Clause *)base;
  if (it->test) {
    base_Print(it->test, stream);
    fputs(" then\n", stream);
    base_Print(it->body, stream);
    if (((Clause *)it->cmd.list.next)->test)
      fputs("elsif ", stream);
    else
      fputs("else\n", stream);
  } else {
    base_Print(it->body, stream);
  }
  if (it->cmd.list.next)
    base_Print(it->cmd.list.next, stream);
}

static void Clause_Eval(Cmd *cmd) {
  Clause *it = (Clause *)cmd;
  if (! it->test || test_Eval(it->test))
    cmd_Eval(it->body);
  else if (it->cmd.list.next)
    cmd_Eval(it->cmd.list.next);
}

static void If_Print(Base *base, FILE *stream) {
  If *it = (If *)base;
  fputs("if ", stream);
  base_Print(it->clause, stream);
  fputs("endif\n", stream);
  if (it->cmd.list.next)
    base_Print(it->cmd.list.next, stream);
}

static void If_Eval(Cmd *cmd) {
  If *it = (If *)cmd;
  cmd_Eval(it->clause);
  if (it->cmd.list.next)
    cmd_Eval(it->cmd.list.next);
}

static void Require_Print(Base *base, FILE *stream) {
  Require *it = (Require *)base;
  fputs("require ", stream);
  fputs(it->ext, stream);
  if (it->cmd.list.next)
    base_Print(it->cmd.list.next, stream);
}

static void Require_Eval(Cmd *cmd) {
  Require *it = (Require *)cmd;
  fprintf(stderr, "Error: Extension \"%s\" not available",
	  it->ext);
  exit(1);
}

static void Bounce_Print(Base *base, FILE *stream) {
  Bounce *it = (Bounce *)base;
  fputs("bounce;\n", stream);
  if (it->cmd.list.next)
    base_Print(it->cmd.list.next, stream);
}

static void Bounce_Eval(Cmd *cmd) {
  Bounce *it = (Bounce *)cmd;
  puts("bounce");
  if (it->cmd.list.next)
    cmd_Eval(it->cmd.list.next);
}  

static void FileInto_Print(Base *base, FILE *stream) {
  FileInto *it = (FileInto *)base;
  fputs("fileinto \"", stream);
  fputs(it->folder, stream);
  fputs("\";\n", stream);
  if (it->cmd.list.next)
    base_Print(it->cmd.list.next, stream);
}

static void FileInto_Eval(Cmd *cmd) {
  FileInto *it = (FileInto *)cmd;
  printf("fileinto %s\n", it->folder);
  if (it->cmd.list.next)
    cmd_Eval(it->cmd.list.next);
}

static void Forward_Print(Base *base, FILE *stream) {
  Forward *it = (Forward *)base;
  fputs("forward \"", stream);
  fputs(it->where, stream);
  fputs("\";\n", stream);
  if (it->cmd.list.next)
    base_Print(it->cmd.list.next, stream);
}

static void Forward_Eval(Cmd *cmd) {
  Forward *it = (Forward *)cmd;
  printf("forward %s\n", it->where);
  if (it->cmd.list.next)
    cmd_Eval(it->cmd.list.next);
}

static void Normal_Print(Base *base, FILE *stream) {
  Normal *it = (Normal *)base;
  fputs("normal;\n", stream);
  if (it->cmd.list.next)
    base_Print(it->cmd.list.next, stream);
}

static void Normal_Eval(Cmd *cmd) {
  Normal *it = (Normal *)cmd;
  puts("normal");
  if (it->cmd.list.next)
    cmd_Eval(it->cmd.list.next);
}

static void Reply_Print(Base *base, FILE *stream) {
  Reply *it = (Reply *)base;
  fputs("reply \"", stream);
  fputs(it->msg, stream);
  fputs("\";\n", stream);
  if (it->cmd.list.next)
    base_Print(it->cmd.list.next, stream);
}

static void Reply_Eval(Cmd *cmd) {
  Reply *it = (Reply *)cmd;
  printf("reply \"%s\"\n", it->msg);
  if (it->cmd.list.next)
    cmd_Eval(it->cmd.list.next);
}

static void Stop_Print(Base *base, FILE *stream) {
  Stop *it = (Stop *)base;
  fputs("stop;\n", stream);
  if (it->cmd.list.next)
    base_Print(it->cmd.list.next, stream);
}

static void Stop_Eval(Cmd *cmd) {
  exit(0);
}

static void Toss_Print(Base *base, FILE *stream) {
  Toss *it = (Toss *)base;
  fputs("toss;\n", stream);
  if (it->cmd.list.next)
    base_Print(it->cmd.list.next, stream);
}

static void Toss_Eval(Cmd *cmd) {
  Toss *it = (Toss *)cmd;
  puts("toss");
  if (it->cmd.list.next)
    cmd_Eval(it->cmd.list.next);
}

/* Tables */

BaseOps StringsOps = {
  &Strings_Print
};

TestOps TestsOps = {
  { &Tests_Print },
  &Tests_Eval
};

TestOps All_OfOps = {
  { &All_Of_Print },
  &All_Of_Eval
};

TestOps Any_OfOps = {
  { &Any_Of_Print },
  &Any_Of_Eval
};

TestOps ExistsOps = {
  { &Exists_Print },
  &Exists_Eval
};

TestOps FalseOps = {
  { &False_Print },
  &False_Eval
};

TestOps HeaderOps = {
  { &Header_Print },
  &Header_Eval
};

TestOps NotOps = {
  { &Not_Print },
  &Not_Eval
};

TestOps SizeOps = {
  { &Size_Print },
  &Size_Eval
};

TestOps SupportOps = {
  { &Support_Print },
  &Support_Eval
};

TestOps TrueOps = {
  { &True_Print },
  &True_Eval
};

CmdOps ClauseOps = {
  { &Clause_Print },
  &Clause_Eval
};

CmdOps IfOps = {
  { &If_Print },
  &If_Eval
};

CmdOps RequireOps = {
  { &Require_Print },
  &Require_Eval
};

CmdOps BounceOps = {
  { &Bounce_Print },
  &Bounce_Eval
};

CmdOps FileIntoOps = {
  { &FileInto_Print },
  &FileInto_Eval
};

CmdOps ForwardOps = {
  { &Forward_Print },
  &Forward_Eval
};

CmdOps NormalOps = {
  { &Normal_Print },
  &Normal_Eval
};

CmdOps ReplyOps = {
  { &Reply_Print },
  &Reply_Eval
};

CmdOps StopOps = {
  { &Stop_Print },
  &Stop_Eval
};

CmdOps TossOps = {
  { &Toss_Print },
  &Toss_Eval
};
